// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
// vi: set et ts=4 sw=4 sts=4:
/*
  This file is part of the Open Porous Media project (OPM).

  OPM is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  OPM is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with OPM.  If not, see <http://www.gnu.org/licenses/>.

  Consult the COPYING file in the top-level source directory of this
  module for the precise wording of the license and the list of
  copyright holders.
*/
/*!
 * \file
 * \copydoc Opm::EnergyModuleType.hpp
 */

#ifndef OPM_ENERGY_MODULE_TYPE_HPP
#define OPM_ENERGY_MODULE_TYPE_HPP

#include <cstdint>

namespace Opm {
    enum class EnergyModules : std::uint8_t {
        NoTemperature = 0,
        ConstantTemperature = 1,
        SequentialImplicitThermal = 2,
        FullyImplicitThermal = 3,
    };
} // namespace Opm

#endif // OPM_ENERGY_MODULE_TYPE_HPP
